/* Copyright (C) 2003, 2004 IRCtoo Network, Pasi Hirvonen <psh@iki.fi>
 *
 * Parts written by FJ (David). Based on the original work by Pasi Hirvonen
 * and Matthew Stephenson. Various others have contributed ideas and testing.*/

/* $Id: magic.c 153 2004-05-25 22:04:30Z pasi $ */

#include <glib.h>
#include <time.h>
#include <string.h>
#include <stdio.h>

#include "spamserv.h"
#include "structs.h"
#include "server.h"
#include "irc.h"
#include "users.h"
#include "ui.h"
#include "log.h"
#include "magic.h"
#include "database.h"
#include "util.h"

GSList *spamstrs;

struct spam_str *spam_str_new(void);
gboolean spam_str_valid(const char *);
int spam_str_cmp(gpointer, gpointer);
struct spam_str *spam_str_find(char *);
int spam_str_idcmp(gpointer, gpointer);
void spam_str_free(struct spam_str *);
void spam_str_clean(gpointer, gpointer);
void spam_str_list_str(gpointer, gpointer);

#define FLAG_TO_HUMAN(u) ((USE_AUTOKILL(u)) ? "autokill"		\
		           : (USE_KILL(u)) ? "kill" : "report")
#define FLAG_CHANNEL(u) ((CHECK_CHANNELS(u)) ? "+channels" : "")

struct spam_searchargs {
	int flags;
	char *string;
	gboolean ignoreflags;
	irc_user *user;
	int showmatchdata;
};

struct spamcheck {
	irc_msg *msg;
	GString *reason;
	char *matchedstr;
	enum actions action;
	int ischan;
};

void s_punish_user(struct spamcheck *sc);
void spam_str_match_string(gpointer str, gpointer spamcheck);
int spam_str_check_spam(struct spamcheck *scheck);

void spam_str_DEV_print(gpointer data, gpointer ignored);
void spam_str_DEV_print_all(void);

/* allocates a new spam_str struct */
struct spam_str *spam_str_new(void)
{
	struct spam_str *new;

	new = g_malloc(sizeof(struct spam_str));
	new->reason = NULL;
	new->comment = NULL;
	new->addedby = NULL;
	new->flags = 0;
	return new;
}

/* free */
void spam_str_free(struct spam_str *str)
{
	if (str->comment != NULL)
		g_string_free(str->comment, TRUE);
	if (str->reason != NULL)
		g_string_free(str->reason, TRUE);
	g_string_free(str->match_str, TRUE);
	g_pattern_spec_free(str->pattern);
	g_free(str->addedby);
	g_free(str);
}

/* Validates the spam string. Our validity checks
 * ensure that too wide spam strings are not added.
 * The string must have less than 50% wildcards and
 * may not be shorter than 7. 7 is the min limit 
 * because of *http* */
gboolean spam_str_valid(const char *s)
{
	int wildcards, len, i;

	wildcards = len = i = 0;
	len = strlen(s);
	for (; i < len; i++)
		 if (s[i] == '*' || s[i] == '?' || s[i] == ' ')
			 wildcards++;
	if (wildcards > len/2)
		return FALSE;
	if (len < 7)
		return FALSE;
	return TRUE;
}

/* compares two spam strings, pointed by p2 and p1
 * using the match string */
int spam_str_cmp(gpointer p1, gpointer p2)
{
	return g_ascii_strcasecmp(((struct spam_str *)p1)->match_str->str,
				(char *)p2);
}

/* compare two spam strings using the id number */
int spam_str_idcmp(gpointer p1, gpointer p2)
{
	return (((struct spam_str *)p1)->id == *(unsigned int *)p2) ? 0 : 1;
}

/* find a spam string and return the pointer, comparison
 * by string. */
struct spam_str *spam_str_find(char *s)
{
	GSList *tmp;

	tmp = g_slist_find_custom(spamstrs, s,
			          (GCompareFunc)spam_str_cmp);
	if (tmp == NULL)
		return NULL;
	return tmp->data;
}

/* same as above, but using id */
struct spam_str *spam_str_find_by_id(unsigned int id)
{
	GSList *tmp;

	tmp = g_slist_find_custom(spamstrs, &id,
				  (GCompareFunc)spam_str_idcmp);
	if (tmp == NULL)
		return NULL;
	return tmp->data;
}

/* 
 * return values:
 *	 1 success, modified flags
 *	 0 success, added string
 *	-1 generic error (missing string or user)
 *	-2 already exists
 *	-3 database error 
 *	-4 invalid spam string
 */	
int spam_str_add(char *str, const char *mask, time_t ttime,
		 int flags, unsigned long id, unsigned long matches,
		 time_t lastmatch, gboolean dbadd, unsigned long *id_r)
{
	struct spam_str *new;
	
	if (str == NULL && mask == NULL)
		return -EGENERICERR;
	strip_all(str);
	if (spam_str_valid(str) == FALSE)
		return -ETOOWIDE;
	/* do we already have such line? */
	if ((new = spam_str_find(str)) != NULL) {
		if (new->flags == flags)
			return -EALREADYEXISTS;
		else {
			new->flags = flags;
			if (database_modify_flags(new, flags) < 0)
			      return -EDBERROR;
			return 1;
		}
	}
	new = spam_str_new();
	new->added = (dbadd == TRUE) ? ttime : time(NULL);
	/* in case of a real add, the function call will have 0s for these. */
	new->last_match = lastmatch;
	new->matches = matches;
	new->addedby = g_strdup(mask);
	new->match_str = g_string_new(str);
	new->pattern = g_pattern_spec_new(str);
	new->flags = (flags) ? flags : SPAMSTR_DEFAULT;
	if (dbadd == TRUE) {
		new->id = id;
	} else {
		if (database_add_spamstr(new) < 0)
			return -3;
		if (id_r != NULL)
			*id_r = new->id;
		i_send_globops("%s added spam line: %s (%s%s)",
			       new->addedby, str, FLAG_TO_HUMAN(new),
			       FLAG_CHANNEL(new));
		s_log(LOG_INFO, "%s added spam line: %s (%s%s)",
		      new->addedby, str, FLAG_TO_HUMAN(new),
		      FLAG_CHANNEL(new));
	}
	spamstrs = g_slist_append(spamstrs, new);
	return 0;
}

int spam_str_add_reason_by_id(unsigned long id, const char *reason)
{
	struct spam_str *tmp;

	if (reason == NULL)
		return -EGENERICERR;
	if ((tmp = spam_str_find_by_id(id)) == NULL)
		return -ENOTFOUND;
	
	return spam_str_add_reason(tmp, reason);
}

int spam_str_add_reason(struct spam_str *str, const char *reason)
{
	if (str == NULL || reason  == NULL)
		return -EGENERICERR;
	if (database_update_reason(str->id, reason) < 0)
		return -EDBERROR;
	if (str->reason)
		str->reason = g_string_assign(str->reason, reason);
	else 
		str->reason = g_string_new(reason);
	return 0;
}

int spam_str_add_comment_by_id(unsigned long id, const char *comment)
{
        struct spam_str *tmp;

        if (comment == NULL)
                return -EGENERICERR;
        if ((tmp = spam_str_find_by_id(id)) == NULL)
                return -ENOTFOUND;

        return spam_str_add_comment(tmp, comment);
}

int spam_str_add_comment(struct spam_str *str, const char *comment)
{
        if (str == NULL || comment  == NULL)
                return -EGENERICERR;
	if (database_update_comment(str->id, comment) < 0)
		return -EDBERROR;
        if (str->comment)
                str->comment = g_string_assign(str->comment, comment);
        else
                str->comment = g_string_new(comment);
        /* copy the database stuff here ;) */
        return 0;
}


/* -1 not found
 * -2 database error */
int spam_str_del_id(unsigned long id)
{
	struct spam_str *tmp;

	if ((tmp = spam_str_find_by_id(id)) == NULL)
		return -ENOTFOUND;
	if (database_del_spamstr(id) < 0)
		return -EDBERROR;
	spamstrs = g_slist_remove(spamstrs, tmp);
	s_log(LOG_INFO, "Removed spam string id %ul", id);
	return 0;
}

void spam_str_DEV_print(gpointer data, gpointer ignored __attr_unused__)
{
	printf("%lu: %s (%s %s)\n",
	      ((struct spam_str *)data)->id,
	      ((struct spam_str *)data)->match_str->str,
	      ((struct spam_str *)data)->addedby,
	      FLAG_TO_HUMAN((struct spam_str *)data));
	      
}

void spam_str_DEV_print_all(void)
{
	g_slist_foreach(spamstrs, (GFunc)spam_str_DEV_print, NULL);
}

void spam_str_list_str(gpointer data, gpointer args)
{
	struct spam_str *str;
	struct spam_searchargs *sa;

	str = data;
	sa = args;
	spam_str_DEV_print(data, args);
printf("%d vs %d\n",  sa->flags, str->flags);
	if (g_pattern_match_simple(sa->string, str->match_str->str) == TRUE)
		if (sa->ignoreflags == TRUE || sa->flags == str->flags) {
			i_send_notice(sa->user, "%lu: %s (%s %s%s)\n",
			      str->id,str->match_str->str, str->addedby,
			      FLAG_TO_HUMAN(str), FLAG_CHANNEL(str));
			if (str->reason != NULL)
				i_send_notice(sa->user, "Reason: %s",
					      str->reason->str);
			if (str->comment != NULL)
				i_send_notice(sa->user, "Comment: %s",
					      str->comment->str);
			if (sa->showmatchdata) {
				i_send_notice(sa->user, "[Last match: %s"
					" (added %s) Total matches: %lu]",
					s_timestamp_to_human(str->last_match),
					s_timestamp_to_human(str->added),
					str->matches);
			}
		}

}

void spam_str_list_to(irc_user *u, int flags, char *s)
{
	struct spam_searchargs searchargs;

	searchargs.string = s;
	if (flags == 0 || flags == SPAMSTR_SHOWINFO)
		searchargs.ignoreflags = TRUE;
	else {
		searchargs.flags = flags;
		searchargs.ignoreflags = FALSE;
	}
printf("XX: %d\n", flags);

	if (SHOW_INFO(flags)) {
		searchargs.showmatchdata = 1;
		searchargs.flags = (flags ^ SPAMSTR_SHOWINFO);
printf("XX: %d\n", flags);
	} else
		searchargs.showmatchdata = 0;
	searchargs.user = u;
	g_slist_foreach(spamstrs, (GFunc)spam_str_list_str, &searchargs);
	i_send_notice(u, "End of spam string list");
}

void spam_str_clean(gpointer data, gpointer ignored __attr_unused__)
{
	spam_str_free((struct spam_str *)data);
}

void spam_str_clean_all(void)
{
	g_slist_foreach(spamstrs, (GFunc)spam_str_clean, NULL);
	g_slist_free(spamstrs);
	spamstrs = NULL;
}

void s_punish_user(struct spamcheck *sc)
{
	/* NEVER NEVER NEVER alloc stuff for outside structs without freeing */
	char *reason;
	int akill_len;
	
	status.stats.matched_spams++;
	if (sc->reason == NULL) {
		reason = g_strdup("This host has been auto-killed for "
				"spamming. You may also be infected with "
				"a trojan virus, please check your "
				"computer.");
	} else {
		reason = strdup(sc->reason->str);
	}
	/* seconds -> minutes */
	akill_len = conf.akilllen * 60;
	(void)s_fill_user_by_nick(sc->msg->user.nick, &sc->msg->user);
	if (sc->action == AUTOKILL) {
		i_send_privmsg(conf.infochan, "autokilled %s for %d minutes.",
				&sc->msg->user.nick, conf.akilllen);
		i_autokill_user(&sc->msg->user, akill_len, reason);
		s_log(LOG_AUTOKILL, "Autokilled %s for %d minutes: %s", 
		      &sc->msg->user.nick, conf.akilllen, sc->matchedstr);
		status.stats.set_autokills++;
	} else if (sc->action == KILL) {
		i_send_privmsg(conf.infochan, "killed %s for spamming: %s",
				&sc->msg->user.nick, sc->matchedstr);
		i_kill_user(&sc->msg->user, reason);
		s_log(LOG_KILL, "Killed %s for spamming: %s.",
		      sc->msg->user.nick, sc->matchedstr);
		status.stats.killed_users++;
	} else {
		if (sc->ischan)
			i_send_privmsg(conf.reportchan,
				"spam: \2%s\2 (%s@%s), matching string: %s (%s)",
				sc->msg->user.nick, sc->msg->user.username,
				sc->msg->user.host, sc->matchedstr,
				sc->msg->target);
		
		else
			i_send_privmsg(conf.reportchan,
				"spam: \2%s\2 (%s@%s): \2%s\2",
				sc->msg->user.nick, sc->msg->user.username,
				sc->msg->user.host, sc->msg->msg);
		i_send_privmsg(conf.reportchan, "autokill: /msg operserv 2 "
				"autokill time 30 *!*@%s %s",
				sc->msg->user.host, reason);
	}
	g_free(reason);
}

void spam_str_do_spamchecks(irc_msg *msg)
{
	struct spamcheck scheck;
	
	strip_all(msg->msg);
	scheck.msg = msg;
	scheck.action = NONE;
	scheck.reason = NULL;
	scheck.ischan = (msg->target[0] == '#') ? 1 : 0;
	s_log(LOG_DEBUG, "Running spamchecks.. ");
	status.stats.parsed_privmsgs++;
	if (spam_str_check_spam(&scheck))
		s_punish_user(&scheck);
	if (scheck.reason)
		g_string_free(scheck.reason, TRUE);
}

void spam_str_match_string(gpointer str, gpointer spamcheck)
{
	struct spamcheck *sc;
	struct spam_str *spamstr;
	irc_msg *msg;
	
	sc = spamcheck;
	spamstr = str;
	msg = sc->msg;
	if (sc->ischan && !CHECK_CHANNELS(spamstr))
		return; 
	if (sc->action == AUTOKILL)
		return;
	if (g_pattern_match_string(spamstr->pattern, msg->msg) == TRUE) {
		/* we set to sc->reason to NULL here if current
		 * string doesn't contain an error message, to prevent
		 * a reason of an another string showing up */
		spamstr->last_match = time(NULL);
		spamstr->matches++;
		database_update_matchdata(spamstr->id, spamstr->matches,
				          spamstr->last_match);
		if (spamstr->reason)
			sc->reason = spamstr->reason;
		else
			sc->reason = NULL;
		sc->matchedstr = spamstr->match_str->str;
		if (USE_AUTOKILL(spamstr)) {
			sc->action = AUTOKILL;
			return;
		}
		if (USE_REPORT(spamstr) && sc->action != KILL) {
			sc->action = REPORT;
		}
		else 
			sc->action = KILL;
	}
}

int spam_str_check_spam(struct spamcheck *scheck)
{
	g_slist_foreach(spamstrs, (GFunc)spam_str_match_string, scheck);
	if (scheck->action != NONE)
		return 1;
	return 0;
}
