/* Copyright (C) 2003, 2004 IRCtoo Network, Pasi Hirvonen <psh@iki.fi>
 *
 * Parts written by FJ (David). Based on the original work by Pasi Hirvonen
 * and Matthew Stephenson. Various others have contributed ideas and testing.*/

/* $Id: server.c 142 2004-05-23 21:14:22Z pasi $ */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <time.h>
#include <glib.h>

#include <sys/uio.h>
#include <unistd.h>

#include "spamserv.h"
#include "structs.h"
#include "irc.h"
#include "server.h"
#include "users.h"
#include "pseudoclient.h"
#include "log.h"
#include "util.h"
#include "cycles.h"

extern GSList *users;

struct cmd cmdtab[] = {
	/* command	usagecnt function to call */
	{"PING",	0,	i_parse_ping},
	{"NICK",	0,	i_parse_nick},
	{"QUIT",	0,	i_parse_quit},
	{"MOTD",	0,	i_parse_motd},
	{"VERSION",	0,	i_parse_version},
	{"STATS",	0,	i_parse_stats},
	{"PRIVMSG",	0,	i_parse_privmsg},
	{"INVITE",	0,	i_parse_invite},
	{"ADMIN",	0,	i_parse_admin},
	{"KICK",	0,	i_parse_kick},
	{"WHOIS",	0,	i_parse_whois},
	{"LUSERS",	0,	i_parse_lusers},
	{"MODE",	0,	i_parse_mode},
	{"SVSMODE",	0,	i_parse_svsmode},
	{"KILL",	0,	i_parse_kill},
	{"INFO",	0,	i_parse_info}, 
	{NULL,		0,	NULL}
};
			
extern struct status_t status;
extern struct conf_t conf;

void s_parse_line(char *);
void s_parse_prefix(char *, irc_msg *);
void s_run_handlers(char *, irc_msg *);
void s_run_scheduling(void);

static GIOChannel *rmtsrv = NULL;

/* 
 * send text to server. return values: 0 (success) -1 (error)
 */
int send_raw(char *fmt, ...)
{
	GIOStatus stat;
	va_list ap;
	char *outgoing;
	gsize i;
	
	if (status.connected == 0)
		return -1;
	va_start(ap, fmt);
	outgoing = g_strdup_vprintf(fmt, ap);
	va_end(ap);
	if (g_str_has_suffix(outgoing, "\n") == FALSE) {
		outgoing=g_realloc(outgoing, strlen(outgoing) + 3);
		strcat(outgoing, "\r\n");
	}
	stat = g_io_channel_write_chars(rmtsrv, outgoing, -1, &i, NULL);
	if (stat == G_IO_STATUS_ERROR) {
		g_free(outgoing);
		return -1;
	}
	if (g_io_channel_flush(rmtsrv, NULL) == G_IO_STATUS_ERROR) {
		g_free(outgoing);
		return -1;
	}
	s_log(LOG_SENT, "%s", outgoing);
	g_free(outgoing);
	return 0;
}

/*
 * connects to remove server, 0 is success, -1 error.
 */
int s_connect_to_server(void)
{
	int remotesocket;
	struct hostent *h;
	struct sockaddr_in dest_addr;
	GError *tmp_error;

	if ((h = gethostbyname(conf.remoteserver)) == NULL) {
		s_log(LOG_ERROR, "Error looking up hostname");
		return -1;
	}
	if ((remotesocket = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		s_log(LOG_ERROR, "Socket error");
		return -1;
	}
	dest_addr.sin_family = AF_INET;
	dest_addr.sin_port = htons(conf.remoteport);
	dest_addr.sin_addr = *((struct in_addr *) h->h_addr);
	memset(&(dest_addr.sin_zero), '\0', 8);
	if (connect(remotesocket, (struct sockaddr *) &dest_addr,
				sizeof(struct sockaddr)) != 0) {
		s_log(LOG_ERROR, "Unable to connect server (connect())!");
		return -1;
	}
	rmtsrv = g_io_channel_unix_new(remotesocket);
	tmp_error = NULL;
	if (g_io_channel_set_encoding(rmtsrv, NULL, &tmp_error)
	    != G_IO_STATUS_NORMAL) {
		if (tmp_error != NULL)
			s_log(LOG_ERROR, "Unable to set encoding: %s\n",
			      tmp_error->message);
		return -1;
	}
	status.connected = 1; 
		
	/* send introduction */
	if (send_raw("PASS %s :TS\r\n", conf.remotepassword) < 0 ||
	    send_raw("CAPAB NOQUIT TSMODE NICKIP") < 0 ||
	    /* the following is really going to be ignored on
	     * bahamut's end now, but hey. for future use. */
	    send_raw("SVINFO 0 0 0 :%ld", time(NULL)) <0 ||
	    send_raw("SERVER %s 1 :%s", conf.servername,
		    conf.serverdesc) < 0) {
		s_log(LOG_ERROR, "Unable to introduce ourselves.");
		return -1;
	}
	status.synched = 0;
	return 0;
}

void s_do_synch(void)
{
	printf("\n\n\n\n");
	printf("SYNCHRONIZATION");
	printf("\n\n\n\n");
	s_introduce_pseudoclients();
	initialize_cycle_checks();
}

/*
 * "main" loop of the program where we actually interact with the
 * server. also runs the schedulers as necessary.
 * */
int s_receive_loop(void)
{
	GString *recvbuf;
	GIOStatus stat;
	gsize term;
	
	recvbuf = g_string_sized_new(1024);
	while (status.quitting != 1) {
		stat = g_io_channel_read_line_string(rmtsrv, recvbuf,
				&term, NULL);
		if (stat == G_IO_STATUS_EOF) {
			g_string_free(recvbuf, TRUE);
			s_log(LOG_ERROR, "Connection closed, EOF");
			s_close_connection("Connection closed, EOF");
			return -ECONNCLOSED;
		}
		if (stat == G_IO_STATUS_ERROR) {
			g_string_free(recvbuf, TRUE);
			s_close_connection("Connection closed, error");
			return -EREADERROR;
		}
		s_log(LOG_RECV, "%s", recvbuf->str);
		s_parse_line(recvbuf->str);
		s_run_scheduling();
	}
	g_string_free(recvbuf, TRUE);
	s_close_connection("Connection to server shut down.");
	return 0;
}

/*
 * close the connection to server and do necessary cleanup
 */
void s_close_connection(const char *msg)
{
	GIOStatus stat;
	
	s_free_slist_generic(&users);
	s_pseudoclient_cleanup();
	/* yes this is bit wrong semantically, it should be s_info in
	 * some cases. shrug. */
	s_log(LOG_ERROR, msg);
	stat = g_io_channel_shutdown(rmtsrv, TRUE, NULL);
	if (stat == G_IO_STATUS_ERROR)
		s_log(LOG_ERROR, "Error in g_io_channel_shutdown()!");
	g_io_channel_unref(rmtsrv);
	status.connected = 0;
}

/*
 * parse message prefix, such as :nick!user@host
 */
void s_parse_prefix(char *s, irc_msg *msg)
{
	char *p;

	if ((p = strchr(s, '@'))) {
		g_strlcpy(msg->user.host, p + 1, MAXHOSTLEN);
		*p = '\0';
	}
	if ((p = strchr(s, '!'))) {
		g_strlcpy(msg->user.username, p + 1, MAXUSERLEN);
		*p = '\0';
	}
	g_strlcpy(msg->user.nick, s + 1, MAXNICKLEN);
}

/*
 * parse a line from server as per rfc1459
 */
void s_parse_line(char *line)
{
	irc_msg msg;
	char **tokens, *p, *command;
	int nexttoken, i;

        /* get rid of \n. we do it early, so we don't have to
	   cut if off twice (for toks and the message */
        if ((p = strchr(line, '\n')) != NULL)
                *p = '\0';
	
	tokens = g_strsplit(line, " ", 0);
	if (tokens == NULL)
		return;
	nexttoken = 0;

	/* a server cannot really send a valid line without a single
	 * space, so if it does.. ignore it for now. */
	if (tokens[0] == NULL || tokens[1] == NULL) {
		s_log(LOG_DEBUG, "Invalid line from server.");
		g_strfreev(tokens);
		return;
	}
	msg.user.nick[0] = msg.user.host[0] = msg.user.username[0] =
		'\0';
        msg.user.gcos[0] = '\0';
	/* do we have a prefix? */
	if (tokens[nexttoken][0] == ':') {
		s_parse_prefix(tokens[nexttoken], &msg);
		nexttoken++;
	}
	if (tokens[nexttoken] == NULL) {
		/* this sure can't be valid */
		g_strfreev(tokens);
		return;
	}
	command = tokens[nexttoken++];
	
	msg.msg = strchr((line[0] == ':') ? line + 1: line, ':');
	
	/* we save the location of msg.msg because we're going to free
	 * it later. the msg.msg pointer is modified by parse_privmsg.
	 * and yes, i do consider this bit hackish. :| */
	msg.msgtoks = tokens + nexttoken;
	
	/* find the length of msgtoks array */
	i = msg.lasttok = 0;
	while (msg.msgtoks[++i] != NULL)
		msg.lasttok++;
	s_run_handlers(command, &msg);
	g_strfreev(tokens);
}

/*
 * execute the handler for each command (PRIVMSG, NOTICE etc)
 */
void s_run_handlers(char *command, irc_msg *msg)
{
	struct cmd *p_cmd;

	for (p_cmd = cmdtab; p_cmd->cmd; p_cmd++) {
		if (strcmp(command, p_cmd->cmd) == 0) {
			p_cmd->usagecount++;
			(void)(*p_cmd->func)(msg);
		}
	}
}

/*
 * run scheduled events (cycle checks and whatnot.
 */
void s_run_scheduling(void)
{
	if (g_timer_elapsed(status.scheduler, 0L) < 5)
		return;
	timeout_run();
	/* reset the scheduler */
	g_timer_start(status.scheduler);
}
